/*
 * QDigiDocCommon
 *
 * Copyright (C) 2012-2013 Jargo Kõster <jargo@innovaatik.ee>
 * Copyright (C) 2012-2013 Raul Metsma <raul@innovaatik.ee>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "QCNG.h"

#include "QPCSC.h"
#include "TokenData.h"

#include <QtCore/QDebug>
#include <QtCore/QLibrary>
#include <QtCore/QStringList.h>
#if QT_VERSION >= 0x050000
#include <QtWidgets/QApplication>
#include <QtWidgets/QWidget>
#else
#include <QtGui/QApplication>
#include <QtGui/QWidget>
#endif
#include <QtNetwork/QSslKey>

#include <qt_windows.h>
#include <WinCrypt.h>
#include <ncrypt.h>

#include <openssl/evp.h>
#include <openssl/rsa.h>

struct QCNGCache
{
	QString guid, provider, key;
	DWORD spec;
};

class QCNGPrivate
{
public:
	typedef SECURITY_STATUS (WINAPI *s_NCryptDecrypt)( NCRYPT_KEY_HANDLE hKey,
		PBYTE pbInput, DWORD cbInput, VOID *pPaddingInfo, PBYTE pbOutput,
		DWORD cbOutput, DWORD *pcbResult, DWORD dwFlags );
	typedef SECURITY_STATUS (WINAPI *s_NCryptEnumKeys)( NCRYPT_PROV_HANDLE hProvider,
		LPCWSTR pszScope, NCryptKeyName **ppKeyName, PVOID *ppEnumState, DWORD dwFlags );
	typedef SECURITY_STATUS (WINAPI *s_NCryptEnumStorageProviders)( DWORD *pdwProviderCount,
		NCryptProviderName **ppProviderList, DWORD dwFlags );
	typedef SECURITY_STATUS (WINAPI *s_NCryptFreeBuffer)( PVOID hObject );
	typedef SECURITY_STATUS (WINAPI *s_NCryptFreeObject)( NCRYPT_HANDLE hObject );
	typedef SECURITY_STATUS (WINAPI *s_NCryptGetProperty)( NCRYPT_HANDLE hObject,
		LPCWSTR pszProperty, PBYTE pbOutput, DWORD cbOutput, DWORD *pcbResult, DWORD dwFlags );
	typedef SECURITY_STATUS (WINAPI *s_NCryptOpenKey)( NCRYPT_PROV_HANDLE hProvider,
		NCRYPT_KEY_HANDLE *phKey, LPCWSTR pszKeyName, DWORD dwLegacyKeySpec, DWORD dwFlags );
	typedef SECURITY_STATUS (WINAPI *s_NCryptOpenStorageProvider)( NCRYPT_PROV_HANDLE *phProvider,
		LPCWSTR pszProviderName, DWORD dwFlags );
	typedef SECURITY_STATUS (WINAPI *s_NCryptSetProperty)( NCRYPT_HANDLE hObject,
		LPCWSTR pszProperty, PBYTE pbInput, DWORD cbInput, DWORD dwFlags );
	typedef SECURITY_STATUS (WINAPI *s_NCryptSignHash)( NCRYPT_KEY_HANDLE hKey,
		VOID *pPaddingInfo, PBYTE pbHashValue, DWORD cbHashValue, PBYTE pbSignature,
		DWORD cbSignature, DWORD *pcbResult, DWORD dwFlags );

	QCNGPrivate();

	void enumKeys( QCNG::Certs &certs, LPCWSTR provider, LPCWSTR scope = 0 );
	NCRYPT_KEY_HANDLE key() const;
	QByteArray prop( NCRYPT_HANDLE handle, LPCWSTR param, DWORD flags = 0 ) const;

	static int rsa_sign( int type, const unsigned char *m, unsigned int m_len,
		unsigned char *sigret, unsigned int *siglen, const RSA *rsa );

	QLibrary lib;
	s_NCryptDecrypt f_NCryptDecrypt;
	s_NCryptEnumKeys f_NCryptEnumKeys;
	s_NCryptEnumStorageProviders f_NCryptEnumStorageProviders;
	s_NCryptFreeBuffer f_NCryptFreeBuffer;
	s_NCryptFreeObject f_NCryptFreeObject;
	s_NCryptGetProperty f_NCryptGetProperty;
	s_NCryptOpenKey f_NCryptOpenKey;
	s_NCryptOpenStorageProvider f_NCryptOpenStorageProvider;
	s_NCryptSetProperty f_NCryptSetProperty;
	s_NCryptSignHash f_NCryptSignHash;

	QCNGCache selected;

	RSA_METHOD method;

	QCNG::PinStatus err;
	QHash<SslCertificate,QCNGCache> cache;
};

QCNGPrivate::QCNGPrivate()
	: lib("ncrypt.dll"), err(QCNG::PinOK)
{
	f_NCryptDecrypt = s_NCryptDecrypt(lib.resolve("NCryptDecrypt"));
	f_NCryptEnumKeys = s_NCryptEnumKeys(lib.resolve("NCryptEnumKeys"));
	f_NCryptEnumStorageProviders = s_NCryptEnumStorageProviders(lib.resolve("NCryptEnumStorageProviders"));
	f_NCryptFreeBuffer = s_NCryptFreeBuffer(lib.resolve("NCryptFreeBuffer"));
	f_NCryptFreeObject = s_NCryptFreeObject(lib.resolve("NCryptFreeObject"));
	f_NCryptGetProperty = s_NCryptGetProperty(lib.resolve("NCryptGetProperty"));
	f_NCryptOpenKey = s_NCryptOpenKey(lib.resolve("NCryptOpenKey"));
	f_NCryptOpenStorageProvider = s_NCryptOpenStorageProvider(lib.resolve("NCryptOpenStorageProvider"));
	f_NCryptSetProperty = s_NCryptSetProperty(lib.resolve("NCryptSetProperty"));
	f_NCryptSignHash = s_NCryptSignHash(lib.resolve("NCryptSignHash"));

	method = *RSA_get_default_method();
	method.name = "QCNG";
	method.rsa_sign = rsa_sign;
}

void QCNGPrivate::enumKeys( QCNG::Certs &certs, LPCWSTR provider, LPCWSTR scope )
{
	QCNGCache c;
	c.provider = QString( (QChar*)provider );

	NCRYPT_PROV_HANDLE h = 0;
	SECURITY_STATUS err = f_NCryptOpenStorageProvider( &h, provider, 0 );

	NCryptKeyName *keyname = 0;
	PVOID pos = 0;
	while( f_NCryptEnumKeys( h, scope, &keyname, &pos, NCRYPT_SILENT_FLAG ) == ERROR_SUCCESS )
	{
		c.key = QString( (QChar*)keyname->pszName );
		c.spec = keyname->dwLegacyKeySpec;
		qWarning() << "key" << c.key
			<< "spec" << c.spec
			<< "alg" << QString( (QChar*)keyname->pszAlgid )
			<< "flags" << keyname->dwFlags;

		NCRYPT_KEY_HANDLE key = 0;
		err = f_NCryptOpenKey( h, &key, keyname->pszName, keyname->dwLegacyKeySpec, NCRYPT_SILENT_FLAG );
		f_NCryptFreeBuffer( keyname );
		keyname = 0;

		SslCertificate cert( prop( key, NCRYPT_CERTIFICATE_PROPERTY ), QSsl::Der );
		c.guid = prop( h, NCRYPT_SMARTCARD_GUID_PROPERTY );
		cache[cert] = c;
		if( !cert.isNull() )
			certs[cert] = cert.type() & SslCertificate::EstEidType || cert.type() & SslCertificate::DigiIDType ?
				c.guid : cert.subjectInfo( QSslCertificate::CommonName );

		f_NCryptFreeObject( key );
	}
	f_NCryptFreeObject( h );
}

NCRYPT_KEY_HANDLE QCNGPrivate::key() const
{
	NCRYPT_PROV_HANDLE prov = 0;
	f_NCryptOpenStorageProvider( &prov, LPCWSTR(selected.provider.utf16()), 0 );
	NCRYPT_KEY_HANDLE key = 0;
	f_NCryptOpenKey( prov, &key, LPWSTR(selected.key.utf16()), selected.spec, 0 );
	f_NCryptFreeObject( prov );
	if( qApp->activeWindow() )
	{
		WId id = qApp->activeWindow()->winId();
		f_NCryptSetProperty( key, NCRYPT_WINDOW_HANDLE_PROPERTY, PBYTE(&id), sizeof(id), 0 );
	}
	return key;
}

QByteArray QCNGPrivate::prop( NCRYPT_HANDLE handle, LPCWSTR param, DWORD flags ) const
{
	if( !handle )
		return QByteArray();
	DWORD size = 0;
	if( f_NCryptGetProperty( handle, param, 0, 0, &size, flags ) )
		return QByteArray();
	QByteArray data( size, 0 );
	if( f_NCryptGetProperty( handle, param, PBYTE(data.data()), size, &size, flags ) )
		return QByteArray();
	return data;
}

int QCNGPrivate::rsa_sign( int type, const unsigned char *m, unsigned int m_len,
	unsigned char *sigret, unsigned int *siglen, const RSA *rsa )
{
	if( type != NID_md5_sha1 && m_len != 36 ) //ssl
		return 0;
	QCNG *d = (QCNG*)RSA_get_app_data( rsa );
	QByteArray sig = d->sign( type, QByteArray( (char*)m, m_len ) );
	if( sig.isEmpty() )
		return 0;
	*siglen = sig.size();
	qMemCopy( sigret, sig.constData(), sig.size() );
	return 1;
}



QCNG::QCNG( QObject *parent )
:	QObject( parent )
,	d( new QCNGPrivate )
{
}

QCNG::~QCNG()
{
	delete d;
}

QByteArray QCNG::decrypt( const QByteArray &data )
{
	d->err = PinUnknown;
	if( !d->f_NCryptDecrypt )
		return QByteArray();

	DWORD size = 256;
	QByteArray res( size, 0 );
	NCRYPT_KEY_HANDLE k = d->key();
	DWORD err = d->f_NCryptDecrypt( k, PBYTE(data.constData()), data.size(), 0,
		PBYTE(res.data()), res.size(), &size, NCRYPT_PAD_PKCS1_FLAG );
	d->f_NCryptFreeObject( k );
	switch( err )
	{
	case ERROR_SUCCESS:
		d->err = PinOK;
		res.resize( size );
		return res;
	case SCARD_W_CANCELLED_BY_USER:
		d->err = PinCanceled; break;
	default: break;
	}
	return QByteArray();
}

QCNG::PinStatus QCNG::lastError() const { return d->err; }

QCNG::Certs QCNG::certs() const
{
	if( !d->f_NCryptEnumStorageProviders )
		return Certs();

	qWarning() << "Start enumerationg providers";
	d->cache.clear();
	Certs result;
	DWORD count = 0;
	NCryptProviderName *names = 0;
	d->f_NCryptEnumStorageProviders( &count, &names, NCRYPT_SILENT_FLAG );
	for( DWORD i = 0; i < count; ++i )
	{
		qWarning() << "Found provider" << QString( (QChar*)names[i].pszName );
		if( wcscmp( names[i].pszName, MS_SMART_CARD_KEY_STORAGE_PROVIDER ) == 0 )
		{
			foreach( const QString &reader, QPCSC().readers() )
			{
				qWarning() << reader;
				QString scope = QString( "\\\\.\\%1\\" ).arg( reader );
				d->enumKeys( result, names[i].pszName, LPCWSTR(scope.utf16()) );
			}
		}
		else
			d->enumKeys( result, names[i].pszName );
	}
	d->f_NCryptFreeBuffer( names );
	qWarning() << "End enumerationg providers";

	return result;
}

Qt::HANDLE QCNG::key()
{
	NCRYPT_KEY_HANDLE k = d->key();
	QSslCertificate cert( d->prop( k, NCRYPT_CERTIFICATE_PROPERTY ), QSsl::Der );
	d->f_NCryptFreeObject( k );
	RSA *rsa = RSAPublicKey_dup( (RSA*)cert.publicKey().handle() );
	if ( !rsa )
		return 0;

	RSA_set_method( rsa, &(d->method) );
	rsa->flags |= RSA_FLAG_SIGN_VER;
	RSA_set_app_data( rsa, this );
	EVP_PKEY *key = EVP_PKEY_new();
	EVP_PKEY_set1_RSA( key, rsa );
	RSA_free( rsa );
	return Qt::HANDLE(key);
}

QStringList QCNG::readers() const
{
	return QPCSC().drivers();
}

TokenData QCNG::selectCert( const SslCertificate &cert )
{
	qWarning() << "Select:" << cert.subjectInfo( "CN" );
	if( !d->f_NCryptFreeObject || !d->cache.contains( cert ) )
		return TokenData();

	d->selected = d->cache[cert];
	qWarning() << "Found:" << d->selected.guid << d->selected.key;
	TokenData t;
	t.setCard( cert.type() & SslCertificate::EstEidType || cert.type() & SslCertificate::DigiIDType ?
		d->selected.guid : cert.subjectInfo( QSslCertificate::CommonName ) );
	t.setCert( cert );

	return t;
}

QByteArray QCNG::sign( int method, const QByteArray &digest ) const
{
	d->err = PinUnknown;
	if( !d->f_NCryptSignHash )
		return QByteArray();

	BCRYPT_PKCS1_PADDING_INFO padInfo;
	padInfo.pszAlgId = 0;
	switch( method )
	{
	case NID_sha1: padInfo.pszAlgId = NCRYPT_SHA1_ALGORITHM; break;
	case NID_sha224: padInfo.pszAlgId = L"SHA224"; break;
	case NID_sha256: padInfo.pszAlgId = NCRYPT_SHA256_ALGORITHM; break;
	case NID_sha384: padInfo.pszAlgId = NCRYPT_SHA384_ALGORITHM; break;
	case NID_sha512: padInfo.pszAlgId = NCRYPT_SHA512_ALGORITHM; break;
	case NID_md5_sha1: //padInfo.pszAlgId = L"SHAMD5"; break;
	default: break;
	}

	DWORD size = 256;
	QByteArray res( size, 0 );
	NCRYPT_KEY_HANDLE k = d->key();
	DWORD err = d->f_NCryptSignHash( k, &padInfo, PBYTE(digest.constData()), digest.size(),
		PBYTE(res.data()), res.size(), &size, BCRYPT_PAD_PKCS1 );
	d->f_NCryptFreeObject( k );
	switch( err )
	{
	case ERROR_SUCCESS:
		d->err = PinOK;
		res.resize( size );
		return res;
	case SCARD_W_CANCELLED_BY_USER:
		d->err = PinCanceled; break;
	default: break;
	}
	return QByteArray();
}
